unit uSynchronizeDirectories;
{$INCLUDE 'DirSync3.inc'}

{.$DEFINE LOG_FILECOPY}
{.$DEFINE LOG_DIRECTORYCOPY}

interface

{$IFNDEF DEBUGLOG}
  {$UNDEF LOG_FILECOPY}
  {$UNDEF LOG_DIRECTORYCOPY}
{$ENDIF}

uses
  SysUtils, Windows, Classes,
  {$IFDEF EUREKALOG} EExceptionManager, {$ENDIF}
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  uSystem, uFiles,
  uDirSyncConsts, uMessages, uSynchronizeDirectoriesDialog, uConfirmations,
  uVisitedDirectories,
  uOptions, uDifferences, uSynchronizeUI, uSupportFunctions;

type
  TSynchronizeTask = class;

  TSynchronizeTask = class(TFarEditorManaged)
  public
    type
      TEditorTask = (
        etNone, etCopyToLeft, etCopyToRight, etCompare,
        etEditLeft, etEditRight, etFileInfo,
        etGotoLeft, etGotoRight,
        etFindNextToLeft, etFindNextToRight
      );
  private
    fLeftDirectory: string;
    fRightDirectory: string;
    fProgress: TFarProgress;
    fCopyFileSourceFileName: string;
    fCopyFileDestinationFileName: string;
    fCopyFileAborted: boolean;
    fSyncUI: TSynchronizeUI;
  protected
    function ProcessKeyEvent(const Info: TKeyEventRecord): boolean; override;
    function GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
    function LoadFiles(out List: TDifferenceList; const FileName: string): boolean;
    procedure ShowProgress(const SourceFileName, DestinationFileName: string; Progress, Max: int64; Msg1, Msg2: TMessages);
    procedure ShowFileCopyProgress(Progress, Max: int64);
    procedure ProgressNeedsRefresh;
    function DeleteDirectoryMissingOnTheOtherPanel(const FileName: string; DeleteOnLeft: boolean): boolean;
    function DeleteFileMissingOnTheOtherPanel(const FileName: string; DeleteOnLeft: boolean): boolean;
    function InternalCopyFileWithProgress(const SourceFileName, DestinationFileName: string; out ErrorCode: DWORD): boolean;
    function CopyFile(const SourceFileName, DestinationFileName: string; FromRightToLeft: boolean): boolean;
    function CopyDirectory(const FileNameWithoutRoot, SourceFileName, DestinationFileName: string; FromRightToLeft: boolean; DirectoriesVisited: TVisitedDirectories): boolean;
    function SyncFile(const SourceFileName, DestinationFileName: string; FromRightToLeft: boolean): boolean;
    function GetFileInfo(out Info: TDifferenceItem): boolean;
    procedure EscapePressedEvent(Sender: TFarProgress; var Value: boolean);
    procedure SetupKeyBar;
    procedure BeforeClose(var CanClose: boolean; var Result: TIntPtr); override;
    procedure AfterShow; override;
    function AskUserWhichTask: TEditorTask;
    function CompareFiles: boolean;
    function FileInfo: boolean;
    function EditFile(IsLeftDirectory: boolean): boolean;
    function GotoFile(IsLeftDirectory: boolean): boolean; overload;
    function ChangeCopyDirection(CopyToLeft: boolean): boolean;
    function FindNextDifference(Differences: TDifferenceTypes): boolean;
    property SyncUI: TSynchronizeUI read fSyncUI;
  public
    class function FindByEditorID(EditorID: TIntPtr; out Editor: TSynchronizeTask): boolean;
    constructor Create(const LeftDirectory, RightDirectory: string);
    destructor Destroy; override;
    function Show(MoveCursorToTheFirstDifference: boolean): boolean;
    function IsEditorLine: boolean;
    function Execute(out ReturnToEditor: boolean): boolean;
    function ExecuteTask: boolean; overload;
    function ExecuteTask(ATask: TEditorTask): boolean; overload;
    procedure Edit(const FileName: string);
    function GotoFile(const FileName: string; OnTheLeft: boolean): boolean; overload;
    procedure VisualCompare(LeftFileName, RightFileName: string);
    property LeftDirectory: string read fLeftDirectory;
    property RightDirectory: string read fRightDirectory;
  end;

implementation

{ TSyncFileInfo }

constructor TSynchronizeTask.Create(const LeftDirectory, RightDirectory: string);
var
  Options: TDirSyncOptions;
  TextStream: TTextStream;
begin
  inherited Create(TempFileName);
  {$IFDEF DEBUGLOG}
  //Log('START TSynchronizeTask.Create("%s", "%s")', [LeftDirectory, RightDirectory]);
  {$ENDIF}
  fLeftDirectory := IncludeTrailingPathDelimiter(LeftDirectory);
  fRightDirectory := IncludeTrailingPathDelimiter(RightDirectory);
  fSyncUI := TSynchronizeUI.Create;
  fProgress := nil;
  Options := TDirSyncOptions.Create;
  try
    Options.Load;
    SyncUI.SilentMode := Options.SilentModeByDefault;
    SyncUI.GlobalRemember := Options.GlobalRemember;
    SyncUI.DisableInternalHotkeys := Options.DisableInternalHotkeys;
    SyncUI.ChangeDirectionInIncompleteBlocks := Options.ChangeDirectionInIncompleteBlocks;
    TextStream := TTextStream.Create(Self.FileName, fmCreate);
    try
      try
        TextStream.Encoding := {$IFDEF UNICODE} feUTF8 {$ELSE} feAnsi {$ENDIF} ;
        TextStream.WriteLine(Format(GetMsg(MReportFileTitle), [RemoveNTSchematicsFromFileName(LeftDirectory), RemoveNTSchematicsFromFileName(RightDirectory)]));
        if Options.ShowDescriptionIsSyncFile then begin
          TextStream.WriteLine(GetMsg(MReportFileNotes1));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF1));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF2));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF3));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF5));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF6));
          TextStream.WriteLine(GetMsg(MReportFileControlsCTRLF1));
          TextStream.WriteLine(GetMsg(MReportFileControlsCTRLF2));
          TextStream.WriteLine(GetMsg(MReportFileControlsCTRLF3));
          TextStream.WriteLine(GetMsg(MReportFileControlsCTRLF5));
          TextStream.WriteLine(GetMsg(MReportFileControlsCTRLF6));
          TextStream.WriteLine(GetMsg(MReportFileControlsALTF1Notes));
          TextStream.WriteLine(GetMsg(MReportFileNotes2));
        end;
        TextStream.WriteLine('');
      finally
        FreeAndNil(TextStream);
      end;
    except
      if FileExists(Self.FileName) then
        SysUtils.DeleteFile(Self.FileName);
      Raise;
    end;
  finally
    FreeAndNil(Options);
    end;
end;

function TSynchronizeTask.ProcessKeyEvent(const Info: TKeyEventRecord): boolean;

  function RunEditorTask(const EditorTask: TEditorTask): boolean;
    begin
      //Log('Running editor task %d', [Integer(EditorTask)]);
      if Self.ExecuteTask(EditorTask) then
      begin
        Result := True;
        Self.Refresh;
        //Log('Running editor task %d -> success', [Integer(EditorTask)]);
      end
      else
      begin
        Result := False;
        //Log('Running editor task %d -> failure', [Integer(EditorTask)]);
      end;
    end;

var
  ShiftFlags: DWORD;
begin
  //inherited;
  Result := False;
  if SyncUI.DisableInternalHotkeys then
    Exit;
  if Info.bKeyDown then
  begin
    ShiftFlags := Info.dwControlKeyState and (LEFT_ALT_PRESSED or LEFT_CTRL_PRESSED or RIGHT_ALT_PRESSED or RIGHT_CTRL_PRESSED or SHIFT_PRESSED);
    //Log('Editor input. Char=%d, VirtualKeyCode=%d, VirtualScanCode=%d, ShiftFlags = %x', [Ord(Info.UnicodeChar), Info.wVirtualKeyCode, Info.wVirtualScanCode, ShiftFlags]);
    if (ShiftFlags = LEFT_ALT_PRESSED) or (ShiftFlags = RIGHT_ALT_PRESSED) then begin
      if Info.wVirtualKeyCode = VK_F1 then
        Result := RunEditorTask(etCopyToLeft)
      else if Info.wVirtualKeyCode = VK_F2 then
        Result := RunEditorTask(etCopyToRight)
      else if Info.wVirtualKeyCode = VK_F3 then
        Result := RunEditorTask(etCompare)
      else if Info.wVirtualKeyCode = VK_F5 then
        Result := RunEditorTask(etFindNextToLeft)
      else if Info.wVirtualKeyCode = VK_F6 then
        Result := RunEditorTask(etFindNextToRight)
    end
    else if (ShiftFlags = LEFT_CTRL_PRESSED) or (ShiftFlags = RIGHT_CTRL_PRESSED) then begin
      if Info.wVirtualKeyCode = VK_F1 then
        Result := RunEditorTask(etEditLeft)
      else if Info.wVirtualKeyCode = VK_F2 then
        Result := RunEditorTask(etEditRight)
      else if Info.wVirtualKeyCode = VK_F3 then
        Result := RunEditorTask(etFileInfo)
      else if Info.wVirtualKeyCode = VK_F5 then
        Result := RunEditorTask(etGotoLeft)
      else if Info.wVirtualKeyCode = VK_F6 then
        Result := RunEditorTask(etGotoRight)
    end
  end;
end;

function TSynchronizeTask.DeleteDirectoryMissingOnTheOtherPanel(const FileName: string; DeleteOnLeft: boolean): boolean;
var
  Attr: DWORD;
  SR: TSearchRec;
  Dir: string;
  ContentDeleted: boolean;
begin
  Result := False;
  ShowProgress(FileName, '', 0, 1, MDeletingDirectory, MNoLngStringDefined);
  //Log('Deleting directory "%s"', [FileName]);
  repeat
    if not GetFileAttributes(FileName, Attr) then begin
      //Log('- File does not exist');
      Result := True;
      Break;
    end
    else begin
      ProgressNeedsRefresh;
      if not SyncUI.AskDeleteConfirmation(FileName, DeleteOnLeft) then begin
        //Log('- User did not allow deleting the destination file');
        Break;
      end;
      // The user allowed me to delete the directory. First delete the content.
      Dir := IncludeTrailingPathDelimiter(FileName);
      //Log('Searching "%s*.*"', [Dir]);
      ContentDeleted := True;
      if (Attr and FILE_ATTRIBUTE_REPARSE_POINT) = 0 then
      begin
        if FindFirst(Dir + '*.*', faAnyFile, SR) = 0 then
          try
            repeat
              //Log('Found item: "%s" (directory=%d)', [SR.Name, Integer((SR.Attr and faDirectory) <> 0)]);
              if (SR.Attr and faDirectory) <> 0 then begin
                if (SR.Name <> '.') and (SR.Name <> '..') then begin
                  if not DeleteDirectoryMissingOnTheOtherPanel(Dir + SR.Name, DeleteOnLeft) then begin
                    ContentDeleted := False;
                    //Log('Failed to delete');
                  end;
                end;
              end
              else begin
                if not DeleteFileMissingOnTheOtherPanel(Dir + SR.Name, DeleteOnLeft) then begin
                  ContentDeleted := False;
                  //Log('Failed to delete');
                end;
              end;
            until FindNext(SR) <> 0;
          finally
            SysUtils.FindClose(SR);
          end;
      end;
      // Then delete the directory itself
      if ContentDeleted then begin
        //Log('Content deleted, deleting "%s"', [FileName]);
        SetFileAttributes(PChar(FileName), Attr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
        if RemoveDirectory(PChar(FileName)) then begin
          //Log('- Deleted successfully');
          Result := True;
          Break;
        end;
      end;
      // If I got here, then the directory couldn't be deleted
      ProgressNeedsRefresh;
      if not SyncUI.AskRetryDelete(MFailedEraseDir, FileName) then
        Break;
    end;
  until False;
  //Log('Result: %d', [Integer(Result)]);
end;

function TSynchronizeTask.DeleteFileMissingOnTheOtherPanel(const FileName: string; DeleteOnLeft: boolean): boolean;
var
  Attr: DWORD;
begin
  Result := False;
  ShowProgress(FileName, '', 0, 1, MDeletingFile, MNoLngStringDefined);
  //Log('Deleting file "%s"', [FileName]);
  repeat
    if not GetFileAttributes(FileName, Attr) then begin
      //Log('- File does not exist');
      Result := True;
      Break;
    end
    else begin
      ProgressNeedsRefresh;
      if not SyncUI.AskDeleteConfirmation(FileName, DeleteOnLeft) then begin
        //Log('- User did not allow deleting the destination file');
        Break;
      end;
      SetFileAttributes(PChar(FileName), Attr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
      if DeleteFile(PChar(FileName)) then begin
        //Log('- Deleted successfully');
        Result := True;
        Break;
      end
      else begin
        ProgressNeedsRefresh;
        if not SyncUI.AskRetryDelete(MFailedEraseFile, FileName) then
          Break;
      end;
    end;
  until False;
  //Log('- Result: %d', [Integer(Result)]);
end;

destructor TSynchronizeTask.Destroy;
begin
  FreeAndNil(fSyncUI);
  FreeAndNil(fProgress);
  if FileName <> '' then
    if FileExists(FileName) then
      SysUtils.DeleteFile(FileName);
  inherited;
end;

procedure TSynchronizeTask.Edit(const FileName: string);
var
  EditorID: TIntPtr;
begin
  TFarEditor.Show(EditorID, FileName, GetMsg(MTitle), 0, 0, -1, -1,
    EF_NONMODAL or EF_IMMEDIATERETURN,
    0, 1, {$IFDEF FAR3_UP} CP_DEFAULT {$ELSE} CP_AUTODETECT {$ENDIF}
  );
end;

function TSynchronizeTask.EditFile(IsLeftDirectory: boolean): boolean;
var
  Info: TDifferenceItem;
  FN: string;
begin
  Result := False;
  if GetFileInfo(Info) then
  begin
    try
      if IsLeftDirectory then
        FN := LeftDirectory + Info.FileName
      else
        FN := RightDirectory + Info.FileName;
      Edit(FN);
      Result := True;
    finally
      FreeAndNil(Info);
    end;
  end;
end;

procedure TSynchronizeTask.EscapePressedEvent(Sender: TFarProgress; var Value: boolean);
begin
  Value := Plugin.ShowMessage([GetMsg(MEscTitle), GetMsg(MEscBody), GetMsg(MOk), GetMsg(MCancel)], FMSG_WARNING, 2) = 0;
end;

function TSynchronizeTask.GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
var
  LastError: DWORD;
  SR: TSearchRec;
begin
  Attributes := Windows.GetFileAttributes(PChar(FileName));
  if Attributes <> INVALID_FILE_ATTRIBUTES then
    Result := True
  else begin
    LastError := GetLastError;
    if (LastError = ERROR_FILE_NOT_FOUND) or (LastError = ERROR_PATH_NOT_FOUND) or (LastError = ERROR_INVALID_NAME) then
      Result := False
    else
      if FindFirst(FileName, faAnyFile, SR) = 0 then
        try
          Attributes := SR.Attr;
          Result := True;
        finally
          SysUtils.FindClose(SR);
        end
      else
        Result := False;
  end;
end;

function TSynchronizeTask.GetFileInfo(out Info: TDifferenceItem): boolean;
var
  s: string;
  EditorInfo: TEditorInfo;
begin
  Result := False;
  if GetInfo(EditorInfo) then
  begin
    if GetString(EditorInfo.CurLine, s) then
    begin
      Info := TDifferenceItem.Create;
      if Info.SetAsString(s) then
        Result := True
      else
      begin
        FreeAndNil(Info);
        Result := False;
      end;
    end;
  end;
end;

function TSynchronizeTask.GotoFile(const FileName: string; OnTheLeft: boolean): boolean;
var
  Panel: TFarPanel;
  Dir, FN: string;
  FileIndex: integer;
  OK: boolean;
begin
  Result := False;
  Panel := TFarPanel.Create(True);
  try
    if Panel.Left <> OnTheLeft then
    begin
      FreeAndNil(Panel);
      Panel := TFarPanel.Create(False);
    end;
    Dir := ExtractFilePath(FileName);
    FN := ExtractFileName(FileName);
    if (Panel.Directory = ExcludeTrailingPathDelimiter(Dir)) or Panel.SetCurrentDir(Dir) then
    begin
      OK := False;
      if FN = '*' then
        OK := True
      else
      if Panel.FindFile(FN, FileIndex) then
      begin
        Panel.CurrentItemIndex := FileIndex;
        if Panel.CurrentItemIndex = FileIndex then
          OK := True;
      end;
      if OK then
      begin
        Panel.Activate;
        Panel.SwitchTo;
        Result := True;
      end;
    end;
  finally
    FreeAndNil(Panel);
  end;
end;

function TSynchronizeTask.GotoFile(IsLeftDirectory: boolean): boolean;
var
  Info: TDifferenceItem;
  FullFileName : string;
begin
  Result := False;
  if GetFileInfo(Info) then
  begin
    try
      if IsLeftDirectory then
        FullFileName := LeftDirectory + Info.FileName
      else
        FullFileName := RightDirectory + Info.FileName;
      if Info.IsDirectory then
        FullFileName := IncludeTrailingPathDelimiter(FullFileName) + '*';
      Result := GotoFile(RemoveNTSchematicsFromFileName(FullFileName), IsLeftDirectory);
    finally
      FreeAndNil(Info);
    end;
  end;
end;

function CopyProgressRoutine(TotalFileSize, TotalBytesTransferred, StreamSize, StreamBytesTransferred: int64; dwStreamNumber, dwCallbackReason: DWORD; hSourceFile, hDestinationFile: THandle; lpData: TSynchronizeTask): DWORD; stdcall;
begin
  Result := PROGRESS_CONTINUE;
  if lpData <> nil then
    try
      lpData.ShowFileCopyProgress(TotalBytesTransferred, TotalFileSize);
    except
      on EAbort do
        begin
          lpData.fCopyFileAborted := True;
          Result := PROGRESS_CANCEL;
        end;
      on Exception do
        Result := PROGRESS_CONTINUE;
    end;
end;

function TSynchronizeTask.InternalCopyFileWithProgress(const SourceFileName, DestinationFileName: string; out ErrorCode: DWORD): boolean;
var
  DestAttr: DWORD;
  SR: TSearchRec;
begin
  {$IFDEF LOG_FILECOPY}
  Log('START InternalCopyFileWithProgress("%s", "%s")', [SourceFileName, DestinationFileName]);
  {$ENDIF}
  if FindFirst(SourceFileName, faAnyFile, SR) = 0 then
  begin
    try
      if GetFileAttributes(DestinationFileName, DestAttr) then begin
        SetFileAttributes(PChar(DestinationFileName), DestAttr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
        DeleteFile(PChar(DestinationFileName));
      end;
      fCopyFileSourceFileName := SourceFileName;
      fCopyFileDestinationFileName := ExtractFilePath(DestinationFileName) + SR.Name;
      fCopyFileAborted := False;
      Result := CopyFileEx(PChar(fCopyFileSourceFileName), PChar(fCopyFileDestinationFileName), @CopyProgressRoutine, Self, nil, 0);
      ErrorCode := GetLastError;
      if Result then
        SetFileAttributes(PChar(DestinationFileName), SR.Attr)
      else if fCopyFileAborted then
        Abort;
    finally
      SysUtils.FindClose(SR);
    end;
  end
  else
    Result := False;
  {$IFDEF LOG_FILECOPY}
  Log('END   InternalCopyFileWithProgress("%s", "%s", %d) -> %d', [SourceFileName, DestinationFileName, ErrorCode, Integer(Result)]);
  {$ENDIF}
end;

function TSynchronizeTask.IsEditorLine: boolean;
var
  Info: TDifferenceItem;
begin
  Result := False;
  if GetFileInfo(Info) then
  begin
    FreeAndNil(Info);
    Result := True;
  end;
end;

function TSynchronizeTask.LoadFiles(out List: TDifferenceList; const FileName: string): boolean;
begin
  Result := False;
  List := TDifferenceList.Create(True);
  try
    List.LoadFromFile(FileName);
    Result := True;
  finally
    if not Result then
      FreeAndNil(List);
  end;
end;

procedure TSynchronizeTask.SetupKeyBar;
{$IFDEF FAR3_UP}
const
  KEY_CTRL_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F3: TFarKey = (VirtualKeyCode: VK_F3; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F5: TFarKey = (VirtualKeyCode: VK_F5; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F6: TFarKey = (VirtualKeyCode: VK_F6; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_ALT_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F3: TFarKey = (VirtualKeyCode: VK_F3; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F5: TFarKey = (VirtualKeyCode: VK_F5; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F6: TFarKey = (VirtualKeyCode: VK_F6; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
{$ENDIF}
var
  {$IFDEF FAR3_UP}
  SetKeyBarTitles: TFarSetKeyBarTitles;
  KeyBarTitles: TKeyBarTitles;
  KeyBarLabels: packed array[0..9] of TKeyBarLabel;
  {$ELSE}
  KeyBarTitles: TKeyBarTitles;
  {$ENDIF}
begin
  if SyncUI.DisableInternalHotkeys then
    Exit;
  {$IFDEF FAR3_UP}
  KeyBarLabels[0].Key := KEY_CTRL_F1;
  KeyBarLabels[0].Text := FarString(GetMsg(MKeyEditLeft));
  KeyBarLabels[0].LongText := FarString(GetMsg(MKeyEditLeftLong));
  KeyBarLabels[1].Key := KEY_CTRL_F2;
  KeyBarLabels[1].Text := FarString(GetMsg(MKeyEditRight));
  KeyBarLabels[1].LongText := FarString(GetMsg(MKeyEditRightLong));
  KeyBarLabels[2].Key := KEY_ALT_F1;
  KeyBarLabels[2].Text := FarString(GetMsg(MKeyOverwriteLeft));
  KeyBarLabels[2].LongText := FarString(GetMsg(MKeyOverwriteLeftLong));
  KeyBarLabels[3].Key := KEY_ALT_F2;
  KeyBarLabels[3].Text := FarString(GetMsg(MKeyOverwriteRight));
  KeyBarLabels[3].LongText := FarString(GetMsg(MKeyOverwriteRightLong));
  KeyBarLabels[4].Key := KEY_ALT_F3;
  KeyBarLabels[4].Text := FarString(GetMsg(MKeyCompare));
  KeyBarLabels[4].LongText := FarString(GetMsg(MKeyCompareLong));
  KeyBarLabels[5].Key := KEY_CTRL_F3;
  KeyBarLabels[5].Text := FarString(GetMsg(MKeyFileInfo));
  KeyBarLabels[5].LongText := FarString(GetMsg(MKeyFileInfoLong));
  KeyBarLabels[6].Key := KEY_ALT_F5;
  KeyBarLabels[6].Text := FarString(GetMsg(MKeyFindNextToLeft));
  KeyBarLabels[6].LongText := FarString(GetMsg(MKeyFindNextToLeftLong));
  KeyBarLabels[7].Key := KEY_ALT_F6;
  KeyBarLabels[7].Text := FarString(GetMsg(MKeyFindNextToRight));
  KeyBarLabels[7].LongText := FarString(GetMsg(MKeyFindNextToRightLong));
  KeyBarLabels[8].Key := KEY_CTRL_F5;
  KeyBarLabels[8].Text := FarString(GetMsg(MKeyGotoLeft));
  KeyBarLabels[8].LongText := FarString(GetMsg(MKeyGotoLeftLong));
  KeyBarLabels[9].Key := KEY_CTRL_F6;
  KeyBarLabels[9].Text := FarString(GetMsg(MKeyGotoRight));
  KeyBarLabels[9].LongText := FarString(GetMsg(MKeyGotoRightLong));
  KeyBarTitles.CountLabels := Length(KeyBarLabels);
  KeyBarTitles.Labels := @KeyBarLabels[0];
  SetKeyBarTitles.StructSize := Sizeof(SetKeyBarTitles);
  SetKeyBarTitles.Titles := @KeyBarTitles;
  FarApi.EditorControl(Self.EditorID, ECTL_SETKEYBAR, 0, @SetKeyBarTitles);
  {$ELSE}
  FillChar(KeyBarTitles, Sizeof(KeyBarTitles), 0);
  KeyBarTitles.CtrlTitles[0] := FarString(GetMsg(MKeyEditLeft));
  KeyBarTitles.CtrlTitles[1] := FarString(GetMsg(MKeyEditRight));
  KeyBarTitles.CtrlTitles[2] := FarString(GetMsg(MKeyFileInfo));
  KeyBarTitles.CtrlTitles[4] := FarString(GetMsg(MKeyGotoLeft));
  KeyBarTitles.CtrlTitles[5] := FarString(GetMsg(MKeyGotoRight));
  KeyBarTitles.AltTitles[0] := FarString(GetMsg(MKeyOverwriteLeft));
  KeyBarTitles.AltTitles[1] := FarString(GetMsg(MKeyOverwriteRight));
  KeyBarTitles.AltTitles[2] := FarString(GetMsg(MKeyCompare));
  KeyBarTitles.AltTitles[4] := FarString(GetMsg(MKeyFindNextToLeft));
  KeyBarTitles.AltTitles[5] := FarString(GetMsg(MKeyFindNextToRight));
  FarApi.EditorControl(ECTL_SETKEYBAR, @KeyBarTitles);
  {$ENDIF}
end;

function TSynchronizeTask.Show(MoveCursorToTheFirstDifference: boolean): boolean;
var
  StartingLine: integer;
  Str: TTextStream;
  Line: string;
  Item: TDifferenceItem;
begin
  Result := inherited Show(GetMsg(MTitle), 0, 0, -1, -1,
        EF_NONMODAL or EF_IMMEDIATERETURN or EF_DELETEONLYFILEONCLOSE or EF_DISABLEHISTORY {$IFDEF FAR3_UP} or EF_DISABLESAVEPOS {$ENDIF} ,
        0, 1, {$IFDEF UNICODE} CP_UTF8 {$ELSE} CP_ANSI {$ENDIF} );
  if Result and MoveCursorToTheFirstDifference then begin
    StartingLine := 0;
    Item := TDifferenceItem.Create;
    try
      Str := TTextStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
      try
        while Str.ReadLine(Line) do begin
          if Item.SetAsString(Line) then begin
            if StartingLine > 0 then begin
              SetCursorPosition(StartingLine);
            end;
            Break;
          end;
          Inc(StartingLine);
        end;
      finally
        FreeAndNil(Str);
      end;
    finally
      FreeAndNil(Item);
    end;
  end;
end;

procedure TSynchronizeTask.ShowFileCopyProgress(Progress, Max: int64);
begin
  ShowProgress(fCopyFileSourceFileName, fCopyFileDestinationFileName, Progress, Max, MCopying, MCopyingTo);
end;

procedure TSynchronizeTask.ShowProgress(const SourceFileName, DestinationFileName: string; Progress, Max: int64; Msg1, Msg2: TMessages);
begin
  if fProgress.NeedsRefresh then begin
    fProgress.Show([
        GetMsg(MTitle)
      , GetMsg(Msg1)
      , fProgress.BuildProgressFilename(RemoveNTSchematicsFromFileName(SourceFileName))
      , GetMsg(Msg2)
      , fProgress.BuildProgressFilename(RemoveNTSchematicsFromFileName(DestinationFileName))
      , fProgress.BuildProgressBar(Progress, Max)
    ]);
  end;
end;

procedure TSynchronizeTask.AfterShow;
begin
  inherited;
  SetupKeyBar;
end;

function TSynchronizeTask.AskUserWhichTask: TEditorTask;
var
  ItemIndex: integer;
begin
  Result := etNone;
  if {TFarPlugin.Instance.}ShowMenu(MENU_EDITOR_GUID, GetMsg(MTitle), '', [
        GetMsg(MOverwriteLeft)
      , GetMsg(MOverwriteRight)
      , GetMsg(MCompare)
      , GetMsg(MEditLeft)
      , GetMsg(MEditRight)
      , GetMsg(MShowFileInfo)
      , GetMsg(MFindNextToLeft)
      , GetMsg(MFindNextToRight)
      , GetMsg(MGotoLeft)
      , GetMsg(MGotoRight)
    ], ItemIndex)
  then
  begin
    case ItemIndex of
      0: Result := etCopyToLeft;
      1: Result := etCopyToRight;
      2: Result := etCompare;
      3: Result := etEditLeft;
      4: Result := etEditRight;
      5: Result := etFileInfo;
      6: Result := etFindNextToLeft;
      7: Result := etFindNextToRight;
      8: Result := etGotoLeft;
      9: Result := etGotoRight;
    end;
  end;
end;

procedure TSynchronizeTask.BeforeClose(var CanClose: boolean; var Result: TIntPtr);
var
  ReturnToEditor: boolean;
begin
  inherited;
  Result := 0;
  Execute(ReturnToEditor);
  if ReturnToEditor then
    CanClose := False;
end;

function TSynchronizeTask.ChangeCopyDirection(CopyToLeft: boolean): boolean;
var
  EditorInfo: TEditorInfo;
  DestDifferenceType: TDifferenceType;
  SrcDifferenceTypes: set of TDifferenceType;
  CurrentLine, SelStart, SelLength, OriginalLine, OriginalTopLine: TIntPtr;
  DiffInfo: TDifferenceItem;
  Line: string;
  IsBlock: boolean;
  CanUndo: boolean;
begin
  //Log('ChangeCopyDirection(%d)', [Integer(CopyToLeft)]);
  Result := False;
  if not GetInfo(EditorInfo) then
    Exit;
  CanUndo := TFarEditor.BeginUndoBlock(EditorID);
  try
    IsBlock := EditorInfo.BlockType <> BTYPE_NONE;
    if CopyToLeft then
    begin
      DestDifferenceType := dtOlderLeft;
      SrcDifferenceTypes := [dtOlderRight, dtDifferent];
      if not IsBlock then
        SrcDifferenceTypes := SrcDifferenceTypes + [dtFailedRight, dtFailedBoth];
    end
    else
    begin
      DestDifferenceType := dtOlderRight;
      SrcDifferenceTypes := [dtOlderLeft, dtDifferent];
      if not IsBlock then
        SrcDifferenceTypes := SrcDifferenceTypes + [dtFailedLeft, dtFailedBoth];
    end;
    //Log('- IsBlock = %d', [Integer(IsBlock)]);
    OriginalTopLine := EditorInfo.TopScreenLine;
    OriginalLine := EditorInfo.CurLine;
    if IsBlock then
      CurrentLine := EditorInfo.BlockStartLine
    else
      CurrentLine := OriginalLine;
    DiffInfo := TDifferenceItem.Create;
    try
      repeat
        //Log('- Loop started');
        //Log('  - CurrentLine = %d', [CurrentLine]);
        if IsBlock then
          if CurrentLine >= EditorInfo.TotalLines then
            Break
          else
          begin
            SetCursorPosition(CurrentLine);
            Inc(CurrentLine);
          end;
        if not GetString(-1, Line, SelStart, SelLength) then
          Break;
        //Log('  - Got string: "%s"', [Line]);
        //Log('  - SelStart = %d, SelLength = %d', [SelStart, SelLength]);
        if IsBlock then
          if (SelLength <= 0) and (Line <> '') then
            Break
          else if SyncUI.ChangeDirectionInIncompleteBlocks then
            // nothing to do
          else if SelStart > 1 then
            Continue
          else
            Line := Copy(Line, SelStart, SelLength);
        //Log('  - Selected string: "%s"', [Line]);
        if DiffInfo.SetAsString(Line) then
          if DiffInfo.DifferenceType in SrcDifferenceTypes then begin
            DiffInfo.DifferenceType := DestDifferenceType;
            DiffInfo.UsedMatch := umUnknown;
            SetString(-1, DiffInfo.AsString);
          end;
        //Log('  - Loop ended');
      until not IsBlock;
    finally
      FreeAndNil(DiffInfo);
    end;
    if IsBlock then
      SetCursorPosition(OriginalLine, -1, -1, OriginalTopLine, -1, -1);
  finally
    if CanUndo then
      TFarEditor.EndUndoBlock(EditorID)
  end;
  //Log('Done');
  Result := True;
end;

function TSynchronizeTask.CompareFiles: boolean;
var
  Info: TDifferenceItem;
begin
  Result := False;
  if GetFileInfo(Info) then
  begin
    try
      VisualCompare(LeftDirectory + Info.FileName, RightDirectory + Info.FileName);
      Result := True;
    finally
      FreeAndNil(Info);
    end;
  end;
end;

function TSynchronizeTask.CopyDirectory(const FileNameWithoutRoot, SourceFileName, DestinationFileName: string; FromRightToLeft: boolean; DirectoriesVisited: TVisitedDirectories): boolean;
const
  ATTRIBUTES_TO_COPY = FILE_ATTRIBUTE_ARCHIVE or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_SYSTEM;
var
  SourceAttr, DestAttr: DWORD;
  SourceDir, DestinationDir: string;
  FileNameWithoutRootDir: string;
  SR: TSearchRec;
  NewDirectoriesVisited: boolean;
begin
  {$IFDEF LOG_DIRECTORYCOPY}
  Log('START CopyDirectory("%s", "%s", %d) START', [SourceFileName, DestinationFileName, Integer(FromRightToLeft)]);
  {$ENDIF}
  Result := False;
  TFarScreen.Title := GetMsg(MTitle) + ': ' + FileNameWithoutRoot;
  ShowProgress(SourceFileName, DestinationFileName, 0, 1, MCopying, MCopyingTo);
  NewDirectoriesVisited := (DirectoriesVisited = nil);
  try
    if NewDirectoriesVisited then
      DirectoriesVisited := TVisitedDirectories.Create;
    if DirectoriesVisited.Visited(SourceFileName) then
    begin
      ShowWarning([GetMsg(MDetectedDirectoryLoop), SourceFileName, GetMsg(MDetectedDirectoryLoop2)]);
      Exit;
    end;
    DirectoriesVisited.Enter(SourceFileName);
    try
      FileNameWithoutRootDir := IncludeTrailingPathDelimiter(FileNameWithoutRoot);
      repeat
        if not GetFileAttributes(SourceFileName, SourceAttr) then begin
          ProgressNeedsRefresh;
          if not SyncUI.AskRetryNotFound(MSourceNotFound, SourceFileName) then
            Break;
        end
        else if (SourceAttr and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
          ProgressNeedsRefresh;
          if not SyncUI.AskRetryNotDirectory(MSourceIsNotDirectory, SourceFileName) then
            Break;
        end
        else begin
          if not GetFileAttributes(DestinationFileName, DestAttr) then begin
            ForceDirectories(DestinationFileName);
            if not GetFileAttributes(DestinationFileName, DestAttr) then begin
              ProgressNeedsRefresh;
              if not SyncUI.AskRetryMakeDir(MCantCreateDirectory, DestinationFileName) then
                Break;
            end;
          end;
          if (DestAttr and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
            ProgressNeedsRefresh;
            if not SyncUI.AskRetryNotDirectory(MDestinationIsNotDirectory, DestinationFileName) then
              Break;
          end;
          //Log('- Copying attributes');
          if (SourceAttr and ATTRIBUTES_TO_COPY) <> (DestAttr and ATTRIBUTES_TO_COPY) then
            SetFileAttributes(PChar(DestinationFileName), (DestAttr and (not ATTRIBUTES_TO_COPY)) or (SourceAttr and ATTRIBUTES_TO_COPY));
          //Log('- Copying files');
          Result := True;
          SourceDir := IncludeTrailingPathDelimiter(SourceFileName);
          DestinationDir := IncludeTrailingPathDelimiter(DestinationFileName);
          if FindFirst(SourceDir + '*.*', faAnyFile, SR) = 0 then
            try
              repeat
                if (SR.Attr and faDirectory) <> 0 then
                  if (SR.Name = '.') or (SR.Name = '..') then
                    Continue
                  else
                    CopyDirectory(FileNameWithoutRootDir + SR.Name, SourceDir + SR.Name, DestinationDir + SR.Name, FromRightToLeft, DirectoriesVisited)
                else
                  CopyFile(SourceDir + SR.Name, DestinationDir + SR.Name, FromRightToLeft);
              until FindNext(SR) <> 0;
            finally
              SysUtils.FindClose(SR);
            end;
          Break;
        end;
      until False;
    finally
      DirectoriesVisited.Leave;
    end;
  finally
    if NewDirectoriesVisited then
      FreeAndNil(DirectoriesVisited);
  end;
  {$IFDEF LOG_DIRECTORYCOPY}
  Log('END   CopyDirectory("%s", "%s", %d) -> %d', [SourceFileName, DestinationFileName, Integer(FromRightToLeft), Integer(Result)]);
  {$ENDIF}
end;

function TSynchronizeTask.CopyFile(const SourceFileName, DestinationFileName: string; FromRightToLeft: boolean): boolean;
var
  SourceAttr, DestAttr, CopyError: DWORD;
  Handle: THandle;
  Msg: TMessages;
begin
  {$IFDEF LOG_FILECOPY}
  Log('START CopyFile("%s", "%s", %d) START', [SourceFileName, DestinationFileName, Integer(FromRightToLeft)]);
  {$ENDIF}
  Result := False;
  ShowProgress(SourceFileName, DestinationFileName, 0, 1, MCopying, MCopyingTo);
  repeat
    if not GetFileAttributes(SourceFileName, SourceAttr) then begin
      ProgressNeedsRefresh;
      if not SyncUI.AskRetryNotFound(MSourceNotFound, SourceFileName) then
        Break;
    end
    else begin
      if GetFileAttributes(DestinationFileName, DestAttr) then
      begin
        ProgressNeedsRefresh;
        if not SyncUI.AskOverwriteConfirmation(SourceFileName, DestinationFileName, FromRightToLeft) then begin
          {$IFDEF LOG_FILECOPY}
          Log('- User did not allow overwriting the destination file');
          {$ENDIF}
          Break;
        end;
      end;
      if InternalCopyFileWithProgress(SourceFileName, DestinationFileName, CopyError) then begin
        {$IFDEF LOG_FILECOPY}
        Log('- Copied successfully');
        {$ENDIF}
        Result := True;
        Break;
      end
      else begin
        {$IFDEF LOG_FILECOPY}
        Log('- Error %d while copying', [CopyError]);
        {$ENDIF}
        Msg := MFailedCopySrcFile;
        Handle := CreateFile(PChar(SourceFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
        if Handle = INVALID_HANDLE_VALUE then
          Msg := MFailedOpenSrcFile
        else begin
          CloseHandle(Handle);
          Handle := CreateFile(PChar(DestinationFileName), GENERIC_WRITE, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
          if Handle = INVALID_HANDLE_VALUE then
            if GetLastError = ERROR_FILE_NOT_FOUND then
              Msg := MFailedCreateDstFile
            else
              Msg := MFailedOpenDstFile
          else
            CloseHandle(Handle);
        end;
        {$IFDEF LOG_FILECOPY}
        Log('- Error cause: %s', [GetMsg(Msg)]);
        {$ENDIF}
				SetLastError(CopyError);
        ProgressNeedsRefresh;
        if not SyncUI.AskRetryCopy(Msg, MFailedCopyDstFile, SourceFileName, DestinationFileName) then
          Break;
      end;
    end;
  until False;
  {$IFDEF LOG_FILECOPY}
  Log('END   CopyFile("%s", "%s", %d) -> %d', [SourceFileName, DestinationFileName, Integer(FromRightToLeft), Integer(Result)]);
  {$ENDIF}
end;

function TSynchronizeTask.Execute(out ReturnToEditor: boolean): boolean;
var
  List: TDifferenceList;
  DiffItem: TDifferenceItem;
  Dlg: TSynchronizeDirectoriesDialog;
  DiffType: TDifferenceType;
  i, Processed, Total: Integer;
  FN: string;
  OK, SyncAgain: Boolean;
  NewTask: TSynchronizeTask;
  DiffFile: TTextStream;
  ScreenTitle: string;
begin
  Result := False;
  ReturnToEditor := True;
  try
    //FarMessage(['DirSync', '', 'Files should be synchronized here:', PChar(FileName), PChar(LeftDirectory), PChar(RightDirectory), '', 'OK'], 0, 1);
    if LoadFiles(List, Self.FileName) then
      try
        if List.Count = 0 then begin
          ReturnToEditor := False;
        end
        else begin
          Dlg := TSynchronizeDirectoriesDialog.Create;
          try
            Dlg.SilentMode := SyncUI.SilentMode;
            Dlg.GlobalRemember := SyncUI.GlobalRemember;
            for DiffType := Low(TDifferenceType) to High(TDifferenceType) do
              Dlg.DiffTypePresent[DiffType] := List.HasDifference[DiffType];
            case Dlg.Execute of
              sdaCancel:
                begin
                  ReturnToEditor := False;
                end;
              sdaSynchronize:
                begin
                  ReturnToEditor := False;
                  //FarMessage(['DirSync', '', 'Files should be synchronized here:', PChar(IntToStr(List.Count) + ' files total'), '', 'OK'], 0, 1);
                  ScreenTitle := TFarScreen.Title;
                  fProgress := TFarProgress.Create;
                  try
                    //fProgress.GranularityMSec := PROGRESS_GRANULARITY_MSEC;
                    fProgress.CheckForEscape := True;
                    fProgress.OnEscapePressed := EscapePressedEvent;
                    SyncUI.Clear;
                    if not Dlg.ConfirmOverwriteLeft then
                      SyncUI[''].OverwriteAction[True, False] := oaYes;
                    if not Dlg.ConfirmOverwriteRight then
                      SyncUI[''].OverwriteAction[False, False] := oaYes;
                    case Dlg.DifferentFilesAction of
                      TSynchronizeDirectoriesDialog.TDifferentFilesAction.dfaSkip:
                        SyncUI[''].DifferentFileAction := dfaSkip;
                      TSynchronizeDirectoriesDialog.TDifferentFilesAction.dfaCopyToLeft:
                        SyncUI[''].DifferentFileAction := dfaLeft;
                      TSynchronizeDirectoriesDialog.TDifferentFilesAction.dfaCopyToRight:
                        SyncUI[''].DifferentFileAction := dfaRight;
                    end;
                    SyncUI.SilentMode := Dlg.SilentMode;
                    SyncUI.GlobalRemember := Dlg.GlobalRemember;
                    try
                      repeat
                        i := 0;
                        Processed := 0;
                        Total := List.Count;
                        while i < List.Count do begin
                          fProgress.SetProgressState( {$IFDEF FAR3_UP} TBPS_NORMAL {$ELSE} PS_NORMAL {$ENDIF} , Processed, Total);
                          DiffItem := List[i];
                          FN := Trim(DiffItem.FileName);
                          TFarScreen.Title := GetMsg(MTitle) + ': ' + FN;
                          OK := True; // If some difference types should be ignored, consider files with that difference successfully processed
                          case DiffItem.DifferenceType of
                            dtMissingLeft:
                              if Dlg.DeleteRightMissingOnLeft then
                                if DiffItem.IsDirectory then
                                  OK := DeleteDirectoryMissingOnTheOtherPanel(RightDirectory + FN, False)
                                else
                                  OK := DeleteFileMissingOnTheOtherPanel(RightDirectory + FN, False)
                              else if Dlg.CopyToLeftPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(FN, RightDirectory + FN, LeftDirectory + FN, True, nil)
                                else
                                  OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, True);
                            dtMissingRight:
                              if Dlg.DeleteLeftMissingOnRight then
                                if DiffItem.IsDirectory then
                                  OK := DeleteDirectoryMissingOnTheOtherPanel(LeftDirectory + FN, True)
                                else
                                  OK := DeleteFileMissingOnTheOtherPanel(LeftDirectory + FN, True)
                              else if Dlg.CopyToRightPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(FN, LeftDirectory + FN, RightDirectory + FN, False, nil)
                                else
                                  OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, False);
                            dtOlderLeft:
                              if Dlg.CopyToLeftPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(FN, RightDirectory + FN, LeftDirectory + FN, True, nil)
                                else
                                  OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, True);
                            dtOlderRight:
                              if Dlg.CopyToRightPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(FN, LeftDirectory + FN, RightDirectory + FN, False, nil)
                                else
                                  OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, False);
                            dtFailedLeft:
                              if Dlg.CopyToLeftPanel then
                                if DiffItem.IsDirectory then
                                  OK := False
                                else
                                  OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, True);
                            dtFailedRight:
                              if Dlg.CopyToRightPanel then
                                if DiffItem.IsDirectory then
                                  OK := False
                                else
                                  OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, False);
                            dtFailedBoth:
                              OK := False;
                            dtDifferent:
                              case Dlg.DifferentFilesAction of
                                TSynchronizeDirectoriesDialog.TDifferentFilesAction.dfaAsk:
                                  begin
                                    ProgressNeedsRefresh;
                                    case SyncUI.AskWhatToDoWithDifferentFiles(LeftDirectory, RightDirectory, FN) of
                                      sfdSkip:
                                        OK := False;
                                      sfdCopyToLeft:
                                        OK := SyncFile(RightDirectory + FN, LeftDirectory + FN, True);
                                      sfdCopyToRight:
                                        OK := SyncFile(LeftDirectory + FN, RightDirectory + FN, False);
                                    end;
                                  end;
                                TSynchronizeDirectoriesDialog.TDifferentFilesAction.dfaSkip:
                                  OK := False;
                                dfaCopyToLeft:
                                  if DiffItem.IsDirectory then
                                    OK := CopyDirectory(FN, RightDirectory + FN, LeftDirectory + FN, True, nil)
                                  else
                                    OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, True);
                                dfaCopyToRight:
                                  if DiffItem.IsDirectory then
                                    OK := CopyDirectory(FN, LeftDirectory + FN, RightDirectory + FN, False, nil)
                                  else
                                    OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, False);
                              end;
                          end;
                          if OK then
                            List.Delete(i)
                          else
                            Inc(i);
                          Inc(Processed);
                        end;
                        SyncAgain := False;
                        if SyncUI.SilentMode then begin
                          //Log('Starting non-silent run');
                          SyncUI.SilentMode := False;
                          if List.Count > 0 then
                            SyncAgain := True;
                        end;
                      until not SyncAgain;
                    except
                      on EAbort do
                        ;
                      on Exception do
                        Raise;
                    end;
                    if List.Count > 0 then begin
                      NewTask := TSynchronizeTask.Create(Self.LeftDirectory, Self.RightDirectory);
                      try
                        DiffFile := TTextStream.Create(NewTask.FileName, fmOpenReadWrite or fmShareDenyWrite);
                        try
                          DiffFile.Position := DiffFile.Size;
                          for i := 0 to Pred(List.Count) do begin
                            DiffFile.WriteLine(List[i].AsString);
                          end;
                        finally
                          FreeAndNil(DiffFile);
                        end;
                        if NewTask.Show(True) then
                          NewTask := nil;
                      finally
                        FreeAndNil(NewTask);
                      end;
                    end;
                  finally
                    FreeAndNil(fProgress);
                    TFarScreen.Title := ScreenTitle;
                  end;
                  TFarPanel.Refresh(True);
                  TFarPanel.Refresh(False);
                  Result := True;
                end;
            end;
          finally
            FreeAndNil(Dlg);
          end;
        end;
      finally
        FreeAndNil(List);
      end;
  except
    on E: EAbort do
      ; //Log('Aborted');
    on E: Exception do
      {$IFDEF EUREKALOG}
      ExceptionManager.ShowLastExceptionData;
      {$ELSE}
      Plugin.ShowMessage([GetMsg(MTitle), E.ClassName, E.Message, GetMsg(MOk)], FMSG_ERRORTYPE or FMSG_WARNING, 1);
      {$ENDIF}
  end;
end;

function TSynchronizeTask.ExecuteTask: boolean;
begin
  Result := ExecuteTask(AskUserWhichTask);
end;

function TSynchronizeTask.ExecuteTask(ATask: TEditorTask): boolean;
begin
  Result := False;
  case ATask of
    etCopyToLeft:      Result := ChangeCopyDirection(True);
    etCopyToRight:     Result := ChangeCopyDirection(False);
    etCompare:         Result := CompareFiles;
    etEditLeft:        Result := EditFile(True);
    etEditRight:       Result := EditFile(False);
    etFileInfo:        Result := FileInfo;
    etFindNextToLeft:  Result := FindNextDifference([dtMissingLeft, dtOlderLeft, dtFailedLeft, dtFailedBoth, dtDifferent]);
    etFindNextToRight: Result := FindNextDifference([dtMissingRight, dtOlderRight, dtFailedRight, dtFailedBoth, dtDifferent]);
    etGotoLeft:        Result := GotoFile(True);
    etGotoRight:       Result := GotoFile(False);
  end;
end;

function TSynchronizeTask.FileInfo: boolean;

  function GetColumns(const Col1, Col2, Col3: string): string;
    begin
      Result := Format('%-20.20s %20.20s %20.20s', [Col1, Col2, Col3]);
    end;

  function GetFileSizeStr(const FileName: string): string;
    var
      Size: int64;
    begin
      Size := GetFileSize(FileName);
      if Size >= 1000000000000 then
        Result := FormatCurr('###,###,##0" MiB"', Size div $100000)
      else if Size >= 0 then
        Result := FormatCurr('###,###,###,##0', Size)
      else
        Result := GetMsg(MUnknown);
    end;

  function GetFileDateStr(Index: integer; const FileName: string): string;
    var
      Dates: array[0..2] of TDateTime;
    begin
      if GetFileDateTime(FileName, Dates[0], Dates[1], Dates[2], True) and (Dates[Index] <> 0) then
        Result := FormatDateTime('dd.mm.yyyy hh:nn:ss', Dates[Index])
      else
        Result := GetMsg(MUnknown);
    end;

var
  Info: TDifferenceItem;
  LeftFN, RightFN: string;
begin
  Result := False;
  if GetFileInfo(Info) then
  begin
    try
      LeftFN := LeftDirectory + Info.FileName;
      RightFN := RightDirectory + Info.FileName;
      ShowMessage([
          GetMsg(MTitle)
        , GetMsg(MShowingFileInfo)
        , TruncateString(Info.FileName, 63, ' ')
        , ''
        , GetColumns('',                            GetMsg(MLeftFile),         GetMsg(MRightFile))
        , GetColumns(GetMsg(MFileSize),             GetFileSizeStr(LeftFN),    GetFileSizeStr(RightFN))
        , GetColumns(GetMsg(MFileCreationTime),     GetFileDateStr(0, LeftFN), GetFileDateStr(0, RightFN))
        , GetColumns(GetMsg(MFileModificationTime), GetFileDateStr(1, LeftFN), GetFileDateStr(1, RightFN))
        , GetColumns(GetMsg(MFileAccessTime),       GetFileDateStr(2, LeftFN), GetFileDateStr(2, RightFN))
        , ''
        , GetMsg(MOk)
        ], 0, 1);
      Result := True;
    finally
      FreeAndNil(Info);
    end;
  end;
end;

class function TSynchronizeTask.FindByEditorID(EditorID: TIntPtr; out Editor: TSynchronizeTask): boolean;
var
  E: TFarEditorManaged;
begin
  Result := False;
  if inherited FindByEditorID(EditorID, E) then
  begin
    if E is TSynchronizeTask then
    begin
      Editor := TSynchronizeTask(E);
      Result := True;
    end;
  end;
end;

function TSynchronizeTask.FindNextDifference(Differences: TDifferenceTypes): boolean;
var
  EditorInfo: TEditorInfo;
  DiffInfo: TDifferenceItem;
  LineNr: integer;
  Line: string;
begin
  Result := False;
  if not GetInfo(EditorInfo) then
    Exit;
  DiffInfo := TDifferenceItem.Create;
  try
    for LineNr := Succ(EditorInfo.CurLine) to Pred(EditorInfo.TotalLines) do
    begin
      if GetString(LineNr, Line) then
      begin
        if DiffInfo.SetAsString(Line) then
        begin
          if DiffInfo.DifferenceType in Differences then
          begin
            SetCursorPosition(LineNr);
            Result := True;
            Break;
          end;
        end;
      end;
    end;
  finally
    FreeAndNil(DiffInfo);
  end;
end;

function TSynchronizeTask.SyncFile(const SourceFileName, DestinationFileName: string; FromRightToLeft: boolean): boolean;
var
  OldAction, NewAction: TRememberedAction;
begin
  //Log('SyncFile("%s", "%s", %d, %d, %d)', [SourceFileName, DestinationFileName, Integer(CanOverwrite), Integer(CanOverwriteReadOnly), Integer(FromRightToLeft)]);
  OldAction := TRememberedAction.Create('');
  try
    NewAction := SyncUI[''];
    OldAction.Assign(NewAction);
    try
      NewAction.OverwriteAction[FromRightToLeft, False] := oaYes;
        // Overwrite is implied - the user already said he wanted to overwrite Left or Right.
        // But that only applies to read-only files.
      Result := Self.CopyFile(SourceFileName, DestinationFileName, FromRightToLeft);
    finally
      NewAction.OverwriteAction[FromRightToLeft, False] := OldAction.OverwriteAction[FromRightToLeft, False];
    end;
  finally
    FreeAndNil(OldAction);
  end;
end;

procedure TSynchronizeTask.ProgressNeedsRefresh;
begin
  if Assigned(fProgress) then
    fProgress.NeedsRefresh := True;
end;

procedure TSynchronizeTask.VisualCompare(LeftFileName, RightFileName: string);
begin
  ProgressNeedsRefresh;
  SyncUI.VisualCompare(LeftFileName, RightFileName);
end;

end.
